export ExportUtils
export @genexport


## Type tools

# basic syntax checker for type expressions to prevent 🥗 in error messages
"""
    check_type(expr)
Check whether `expr` is a syntactically valid type specification.

!!! warning "Do NOT assume this is secure!"

    This function is by no means perfect and may occasionally fail!
    Do **NOT NOT NOT** use it in security-critical contexts unless you understand every
    aspect of it internals.

# Extended help

Returns false unless one of the following conditions is true
 1. `expr` is a `Symbol`
 2. `expr` consists only of `:(.)` expressions and `Symbol`s
 3. `expr` is a `:curly` expression, with the first argument matching conditions 1 and 2
    and the remaining arguments are valid type parameters, defined as being either `Symbol`s,
    `isbits` literals or `:(<:)`  expressions, in which `check_type` is true for the last
    argument and all additional arguments are `Symbol`s
 4. `expr` is a `:where` expression with the first argument fulfilling condition 3 and all
    remaining arguments are valid type parameters as defined in condition 3

If a `QuoteNode` is encountered in any of the recursive steps, the condition is applied to
its contents.
"""
function check_type(t)
    if !_check_type(t)
        throw(ArgumentError(("Not a type: $t")))
    end
    return true
end

# reject all expressions except for a.b and a{b}
_check_type(q::QuoteNode) = _check_type(q.value)
function _check_type(t)
    if _check_simple_type(t)
        return true
    elseif t.head == :curly
        return _check_simple_type(t.args[1]) && all(map(_check_params, t.args[2:end]))
    elseif t.head == :where
        return (t.args[1].head == :curly && _check_type(t.args[1])) && all(map(_check_params, t.args[2:end]))
    elseif t.head == :escape
        return _check_type(only(t.args))
    end
    return false
end


_check_simple_type(::Symbol) = true
_check_simple_type(t::Expr) = t.head == :(.) && all(map(_check_simple_type, t.args))
_check_simple_type(q::QuoteNode) = _check_simple_type(q.value)


# _check_params returns true for Symbols, valid subtyping expressions and `isbits` literals
_check_params(::Symbol) = true
_check_params(q::QuoteNode) = _check_params(q.value)
function _check_params(p::Expr)
    if (p.head == :(<:))
        if length(p.args) > 1
            return p.args[1] isa Symbol && _check_type(p.args[2])
        else
            return _check_type(p.args[])
        end
    end
    if (p.head == :tuple)
        return all(map(_check_type, p.args))
    end
    return _check_type(p)
end

_check_params(p) = isbits(p)



"""
    splitstoredtype(t)
Splits an expression representing a type into type name, type parameters and placeholder symbols
for omitted parameters.

## Example
```julia
splitstoredtype(:(Foo{Bar, Baz} where Baz))
```
will return
```
:Foo, Any[:Bar, :Baz], Any[:Baz]
```
"""
function splitstoredtype(t::Expr)
    omitted = []
    if t.head == :where
        omitted = t.args[2:end]
        t = t.args[1]
    end
    @assert t.head == :curly
    return t.args[1], t.args[2:end], omitted
end

function splitstoredtype(t::Symbol)
    return t, [], []
end


"""
    freeparams(T::Type)
Returns the number of free type parameters.
"""
function freeparams(T::UnionAll)
    i = 0
    while T isa UnionAll
        T = T.body
        i += 1
    end
    i
end
freeparams(::Type) = 0




"""
    reconstruct_type(name::String; [mod = Main])
Reconstruct the type represented by `name`, by parsing it and evaluating the resulting
expression in `mod`

Uses [`check_type`](@ref) to prevent the most basic of code injection attacks. Yay security!
"""
function reconstruct_type(name::String; mod::Module = Main)
    expr = Meta.parse(name)
    check_type(expr)
    t = try
        mod.eval(expr)
    catch
        throw(ArgumentError("Could not reconstruct $name in module $(mod)"))
    end
    @assert t isa Type "Not a type: $name"
    t
end


"""
    reconstruct_subtype(name::String, super::Type; [mod = Main])
Same as [`reconstruct_type`](@ref), but complains loudly if the reconstructed type is not a
descendant of `super`.
"""
function reconstruct_subtype(name::String, super::Type; kwargs...)
    T = reconstruct_type(name; kwargs...)
    if !(T<:super)
        error("$name could be reconstructed, but is not a subtype of $super.")
    end
    return T
end


## Export dispatch system


"""
    ExportItem
Valtype for dispatching over export items
"""
abstract type ExportItem{S} end

"""
    FieldExport(S::Symbol)
[`ExportItem`](@ref) representing export of a struct field
"""
struct FieldExport{S} <: ExportItem{S}
    FieldExport(s::Symbol) = new{s}()
end

"""
    AdditionalExport(i::Int)
[`ExportItem`](@ref) representing additional exports (as defined by [`@exportadd`](@ref))
"""
struct AdditionalExport{S} <: ExportItem{S}
    AdditionalExport(i::Integer) = new{i}()
end

#%%

"""
    ExportStyle(T)
Trait which determines how the export system treats instances of `T`.
Defaults to [`VectorExport`](@ref) for subtypes of [`AbstractParticle`]
and [`ScalarExport`](@ref) for everything else.
"""
abstract type ExportStyle  end

"""
    ScalarExport()
[`ExportStyle`](@ref) for types that should be serialized individually, i.e. only one instance
should be written into a data file group.
"""
struct ScalarExport <: ExportStyle end

"""
    VectorExport()
[`ExportStyle`](@ref) for types that should be serialized as vectors, i.e. many instances
of the type should be written into the same group, with fields being collected into vectors.

Note that some backends may not like vectors of complex field types, so be sure to
use [`@exportcollect`](@ref) to pre-digest them with a custom collection function.
"""
struct VectorExport <: ExportStyle end

# Defaults
ExportStyle(::Type) = ScalarExport()
ExportStyle(::Type{<:AbstractParticle}) = VectorExport()


"""
    ExportUtils
This module serves as a namespace for all functions generated by the import/export system.
"""
module ExportUtils
    using InPartS
    import InPartS:SV, ExportItem, FieldExport, AdditionalExport

    """
        exportfield(::Type, ::ExportItem)
    If false, the given [`ExportItem`] is skipped during export. Used for
    [`AdditionalExport`](@ref) to determine the last additional export.
    """
    exportfield(::Type, ::FieldExport) = true
    exportfield(::Type, ::AdditionalExport) = false


    # default functions for things
    exporttransform(::Type, ::ExportItem) = :(value)
    importtransform(::Type, ::ExportItem) = :(data)

    exportcollect(::Type, ::ExportItem) = :(vector)
    importcollect(::Type, ::ExportItem) = :(blob)

    # defaults for params: use the fact the the simulation will be available in the (de)serializers
    exporttransform(::Type{<:AbstractParticle}, ::FieldExport{:params}) = :(findfirst(x->x==value, sim.params))
    importtransform(::Type{<:AbstractParticle}, ::FieldExport{:params}) = :(sim.params[data])

    # defaults for centerpos: transform  into 2D array for export
    exportcollect(::Type{<:AbstractParticle}, ::FieldExport{:centerpos}) = :(InPartS.collectvectors(vector))
    importcollect(::Type{<:AbstractParticle}, ::FieldExport{:centerpos}) = :([InPartS.SVector{size(blob)[2], Float64}(blob[i,:]) for i ∈ 1:size(blob)[1]]) #TODO: faster‽

    # default names (trivial for FieldExport, irrelevant for AdditionalExport)
    exportname(::Type, ::FieldExport{S}) where S = string(S)
    exportname(::Type, ::AdditionalExport) = nothing
    importname(::Type, ::FieldExport{S}) where S = S # if you value your life, do NOT overwrite this method!!!!!
    importname(t::Type, a::AdditionalExport) = Symbol(exportname(t, a))

    #modname(::Type) = "Main"
    # more defaults!
    """
        serialize!(group, data::Vector{<:AbstractParticle}; [sim::Simulation])
        serialize!(group, data::AbstractParams)
    Serialize the `data` into `group`.

    For more information on the export system, consult the InPartS manual.
    """
    serialize!(g, d::T; kwargs...) where {T} = serialize!(g, d, InPartS.ExportStyle(T); kwargs...)
    serialize!(::Any, @nospecialize(a), @nospecialize(s::InPartS.ExportStyle); kwargs...) = error("InPartS: no $s serializer defined for type $(typeof(a))")

    """
        deserialize(group, type::Type{T}; [sim::Simulation]) where {T<:AbstractParticle} → particles::Vector{T}
        deserialize(group, type::Type{T}) where {T<:AbstractParams} → param::T
    Deserializes the `group` into either a vector of `type` or a single instance of `type`.

    For more information on the export system, consult the InPartS manual.
    """
    deserialize(g, t; kwargs...) = serialize!(g, d, InPartS.ExportStyle(t); kwags...)
    deserialize(::Any, @nospecialize(a::Type), @nospecialize(s::InPartS.ExportStyle); kwargs...) = error("InPartS: no $s deserializer defined for type $a")
end

"""
    @exportoff(type, fieldname)
Disable export the field `fieldname` for all instances of `type`.

!!! note
    Be aware that to be able to import data with disable exports, you need to
    provide a kwargs-only constructor where `fieldname` is not a required argument.
"""
macro exportoff(type, field::Symbol)
    check_type(type)
    val = FieldExport{field}
    return quote
        ExportUtils.exportfield(::Type{<:$(esc(type))}, ::$val) = false
    end
end


"""
    @exporttransform(type, fieldname, fwd, back)
Provide expressions to transform the field `fieldname` for all instances of `type`
for (de)serialization.

### Arguments
*   `fwd` A non-quoted expression specifying how to transform the exported value.
    In the context of this expression, the symbol `value` represents field value
    of the instance which is being exported.
*   `back` A non-quoted expression specifying how to reverse-process the
    information during import. The symbol `data` in this expression represents
    the imported data.

!!! note
    This macro only works for fields, not for manually added exports
"""
macro exporttransform(type, field::Symbol, fwd, back)
    check_type(type)
    val = FieldExport{field}
    return quote
        ExportUtils.exporttransform(::Type{<:$(esc(type))}, ::$val) = $(QuoteNode(fwd))
        ExportUtils.importtransform(::Type{<:$(esc(type))}, ::$val) = $(QuoteNode(back))
    end
end

"""
    @exportrecursive(type, fieldname)
Convenience wrapper for [`@exporttransform`](@ref) indicating that the field `fieldname`
of `type` is itself an InPartS-exportable type (produced, e.g., with @genexport).

This macro produces code
```julia
@exporttransform {type} {fieldname} ExportUtils.serialize!(Dict{String, Any}("_type" => string(typeof(value))), value) ExportUtils.deserialize(data, InPartS.reconstruct_subtype(data["_type"], {fieldtype}))
```
where `fieldtype` is the automatically determined type of the field.
"""
macro exportrecursive(type, field::Symbol)
    #check_type(fieldtype)
    check_type(type)
    print(parentmodule(__module__.eval))
    local ft = fieldtype(__module__.eval(type), field)
    println(ft)
    return quote
        @exporttransform $(esc(type)) $(field) ExportUtils.serialize!(Dict{String, Any}("_type" => string(typeof(value))), value) ExportUtils.deserialize(data, InPartS.reconstruct_subtype(data["_type"], $(ft)))
    end
end


"""
    @exportcollect(type,fieldname,expr,backexpr)
Provide expressions for converting the accumulated `fieldname` values into an
exportable object (e.g an array).

### Arguments
*   `expr` A non-quoted expression specifying how to transform the export.
    In the context of this expression, the symbol `vector` represents the vector
    of `field` values for all the instances of type `type` in the collection
    (for example a particle vector).
*   `backexpr` A non-quoted expression specifying how to reverse-process the
    information during import. The symbol `blob` in this expression represents
    the imported data.

!!! note
    This macro only works for [`VectorExport`](@ref) types!
"""
macro exportcollect(type, field::Symbol, fwd, back)
    check_type(type)
    fieldname = String(field)
    return quote
        let f = ($fieldname), T = $(esc(type))
            fs = Symbol(f)
            val = nothing

            if fs ∈ fieldnames(T)
                val = InPartS.FieldExport{fs}
            else

                local i = 1
                while Base.invokelatest(InPartS.ExportUtils.exportfield, T, InPartS.AdditionalExport(i))
                    if Base.invokelatest(InPartS.ExportUtils.exportname, T, InPartS.AdditionalExport(i)) == f
                        val = AdditionalExport{i}
                        break
                    end
                    i = i + 1
                end
                isnothing(val) && error("Export name $f not found!")
           end

           ExportUtils.exportcollect(::Type{<:$(esc(type))}, ::val) = $(Expr(:quote, fwd))
           ExportUtils.importcollect(::Type{<:$(esc(type))}, ::val) = $(QuoteNode(back))
        end
    end
end

"""
    @exportadd type name fwd back
Adds an additional export field of name `name` to `type`.

### Arguments
 * `fwd` A non-quoted expression specifying how to compute the value of the additional
    export from an instance of `type`.
    In the context of this expression, the symbol `item` represents the instance
    of `type`.
*  `back` A non-quoted expression specifying how to process the information during
    import.
    The symbol `data` in this expression represents the imported data.
    The result of this transformation will be passed to the kwarg-only constructor
    of `type` when reconstructing the data.

!!! warning

    If there is no kwargs-only constructor of your type that accepts `name` as a keywoard argument,
    importing your data will fail!
"""
macro exportadd(type, newname, fwd, back)
    check_type(type)

    name = string(newname)
    return quote
        let name = $name, T = $(esc(type))
            if Symbol(name) ∈ fieldnames(T)
                error("Export name $name already exists, use @exporttransform to transform an existing export name")
            else
                local i = 1
                while Base.invokelatest(InPartS.ExportUtils.exportfield, T, InPartS.AdditionalExport(i))
                    if Base.invokelatest(InPartS.ExportUtils.exportname, T, InPartS.AdditionalExport(i)) == name
                        @warn "additional export $name already exists, overwriting"
                        break
                    end
                    i = i + 1
                end

                val = AdditionalExport{i}
                ExportUtils.exporttransform(::Type{<:T}, ::val) = $(QuoteNode(fwd))
                ExportUtils.importtransform(::Type{<:T}, ::val) = $(QuoteNode(back))
                ExportUtils.exportname(::Type{<:T}, ::val) = $(name)
                ExportUtils.exportfield(::Type{<:T}, ::val) = true
            end
        end
    end
end

"""
    @genexport(type<:AbstractParticle)
Generates the import/export functions for the given particle and its [`InPartS.ParamType`](@ref).

!!! warning

    This macro evals generated code into `InPartS.ExportUtils`. Using it as a top-level
    statement in a precompiled package will bring naught but pain and destruction.

    If you want to automatically generate export/import functions when loading a package,
    consider putting `@genexport` in the [`__init__()`](@ref) function.
"""
macro genexport(type::Union{Symbol,Expr})
    return quote
        local mod = @__MODULE__
        Base.eval(mod, _serialize_function($(esc(type))))
        Base.eval(mod, _deserialize_function($(esc(type))))
        #ExportUtils.modname(::Type{mod.$type}) = "$mod"
    end
end



## Serialization


"""
    _serialize_function(T)
Returns an expression containing the definition of [`ExportUtils.serialize!()`](@ref)
for the given type `T`.

The signature of `serialize!` may vary depending on whether `T` is an
[`AbstractParticle`](@ref) or [`AbstractParams`](@ref). For more information on
the export system, consult the InPartS manual.
"""
_serialize_function(T::Type) = _serialize_function(T, ExportStyle(T))

_serialize_function(T::Type, es::ScalarExport) = quote
    function ExportUtils.serialize!(group, item::$T)
        $(_serialize_block(T, es))
        return group
    end
end

_serialize_function(T::Type, es::VectorExport) = quote
    function ExportUtils.serialize!(group, items::Vector{<:$T}; sim::Union{Simulation, Nothing} = nothing)
        $(_serialize_block(T, es))
        return group
    end
end


"""
    _serialize_block(T)
Returns an expression containing the interior of  [`ExportUtils.serialize!()`](@ref)
for the given type `T`.

For more information on the export system, consult the InPartS manual.
"""
_serialize_block(T::Type, es::ExportStyle) = Expr(:block, [_serialization_expr(T, es, e) for e ∈ allexports(T)]...)


"""
    _serialization_expr(T, e, [name, transform, collect])
Returns the expression for serializing the [`ExportItem`](@ref) `e` for the type `T`

Optionally specify custom expressions for export name, transformation and collection.
The collection expression will be ignored for `ScalarExport`. For more
information on the export system, consult the InPartS manual.
"""
function _serialization_expr(T::Type, es::ScalarExport, e::ExportItem)
    name = ExportUtils.exportname(T, e)
    transform = ExportUtils.exporttransform(T, e)
    _serialization_expr(T, es, e, name, transform)
end

function _serialization_expr(T::Type, es::VectorExport, e::ExportItem)
    name = ExportUtils.exportname(T, e)
    transform = ExportUtils.exporttransform(T, e)
    collect = ExportUtils.exportcollect(T, e)
    _serialization_expr(T, es, e, name, transform, collect)
end

# params, field export —> access & transform
_serialization_expr(::Type, ::ScalarExport, ::FieldExport{F}, name, transform,) where F = quote
    group[$name] = begin
        value = item.$F
        $transform
    end
end

# params, add. export —> transform
_serialization_expr(::Type{<:AbstractParams}, ::ScalarExport, ::AdditionalExport, name, transform) = quote
    group[$name] = $transform
end

# particle, field export —> access, transform & collect
_serialization_expr(T::Type, ::VectorExport, ::FieldExport{F}, name, transform, collect) where F = quote
    group[$name] = begin
        vector = map(items) do item::$T
            value = item.$F
            return $transform
        end
        blob = $collect
    end
end

# particle, add. export —> transform & collect
_serialization_expr(::Type, ::VectorExport, ::AdditionalExport, name, transform, collect) = quote
    group[$name] = begin
        vector = map(item->$transform, items)
        blob = $collect
    end
end



## Deserialization
"""
    _deserialize_function(T)
Returns an expression containing the definition of [`ExportUtils.deserialize()`](@ref)
for the given type `T`.

The return type of `deserialize` may vary depending on the type's [`ExportStyle`](@ref). For more information on
the export system, consult the InPartS manual.
"""
_deserialize_function(T::Type) = quote
    function ExportUtils.deserialize(group, type::Type{<:$T}; sim::Union{Simulation, Nothing} = nothing)
        $(_deserialize_block(T))
        return items
    end
end

"""
    _deserialize_block(T)
Returns an expression containing the interior of  [`ExportUtils.deserialize()`](@ref)
for the given type `T`.

For more information on the export system, consult the InPartS manual.
"""
_deserialize_block(T::Type) = _deserialize_block(T, ExportStyle(T))

function _deserialize_block(T::Type, ::ScalarExport)
    datakwargs = Expr(:parameters)

    for e ∈ allexports(T)
        iname = ExportUtils.importname(T, e)
        source = :(group[$(ExportUtils.exportname(T, e))])
        transform = ExportUtils.importtransform(T, e)

        push!(datakwargs.args, Expr(
            :kw,
            iname,
            quote
                data = $source
                $transform
            end
        ))

    end
    constructor = Expr(:call, :type, datakwargs)
    return :(items = $constructor)
end

function _deserialize_block(T::Type, ::VectorExport)
    datakwargs = Expr(:parameters)
    inames = Expr(:parameters)

    for e ∈ allexports(T)
        iname = ExportUtils.importname(T, e)
        source = :(group[$(ExportUtils.exportname(T, e))])
        collect = ExportUtils.importcollect(T, e)
        transform = ExportUtils.importtransform(T, e)

        push!(datakwargs.args, Expr(
            :kw,
            iname,
            quote
                blob = $source
                datavec = $collect
                values = map(data->$transform, datavec)
            end
        ))

        push!(inames.args, Expr(:kw, iname, :(data.$iname[i])))
    end

    datatuple = Expr(:tuple, datakwargs)
    return quote
        # all the importing happens here!
        data = $datatuple
        nitems = length(data[1])
        items = Vector{type}(undef, nitems)
        for i ∈ 1:nitems
            items[i] = type($inames)
        end
    end
end


## Utility functions

"""
    fieldexports(T)
collects all field exports of a type `T`
"""
function fieldexports(T::Type)
    exports = FieldExport[]
    sizehint!(exports, fieldcount(T))
    for fn ∈ fieldnames(T)
        e = FieldExport(fn)
        if ExportUtils.exportfield(T, e)
            push!(exports, e)
        end
    end
    return exports
end

"""
    additionalexports(T)
collects all addditional exports of a type `T`
"""
function additionalexports(T::Type)
    exports = AdditionalExport[]
    i = 1
    while true
        e = AdditionalExport(i)
        if !ExportUtils.exportfield(T, e)
            break
        end
        push!(exports, e)
        i += 1
    end
    return exports
end

"""
    allexports(T)
collects all exports of a type `T`
"""
allexports(T::Type) = vcat(fieldexports(T), additionalexports(T))

"""
    collectvectors(vs::AbstractVector{SVector{D, T}}) where {D, T}
Collect `N` `D`-dimensional vectors into an `N×D` matrix.
"""
function collectvectors(vs::AbstractVector{SVector{D, T}}) where {D, T}
    out = Matrix{T}(undef, length(vs), D)

    for (i, vs) ∈ enumerate(vs)
        for j ∈ 1:D
            out[i, j] = vs[j]
        end
    end

    return out
end
